#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 22 10:44:55 2023

@author: chris
"""

import numpy as _np
from ..math import _rotate_2d


def gaussian(res, A0, sigma, dark=0, center=None, order=2, theta=0):
    '''
    Generate an image containing a Gaussian beam, e.g. in the simplest case
    image = A0 * exp(-2*(r/2sigma)^2)

    Parameters
    ----------
    res : int or [int, int]
        resolution of image panel
    A0 : float
        peak amplitude
    sigma : float or [float, float]
        width of Gaussian w=2sigma
    dark : float
        dark level. The default is 0.
    center : (x, y), optional
        pixel of center of Gaussian. The default is None, meaning res/2.
    order : int
        exponent order of Gaussian function. The default is 2.
    theta : float, optional
        if sigma is paired, this is the rotation of sigma, so that an
        assymmetric Gaussian can be at any angle. Default is 0

    Returns
    -------
    image : 2d array of float
        image of Gaussian
    '''
    res *= _np.array([1, 1], dtype=int)
    if center is None:
        center = res / 2
    cx, cy = tuple(center)
    #x, y = _np.meshgrid(_np.arange(res[0]), _np.arange(res[1]))
    x, y = _np.meshgrid(_np.arange(res[0]) - cx, _np.arange(res[1]) - cy)
    if not _np.isscalar(sigma) and not _np.isclose(theta, 0):
        x, y = _rotate_2d(x, y, theta)
    w = sigma * _np.array([2, 2], dtype=int)
    image = dark + A0 * _np.exp(
        -2*((x/w[0])**order + (y/w[1])**order))
    return image



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb 12 20:24:29 2025

@author: chris
"""

import numpy as _np
from skimage import morphology as _morph

from .roi import Regions


def ringRegions(roi, pad, *args, exclusion=None):
    '''
    For any Regions object, expand forming a ring around the orginal.

    Parameters
    ----------
    roi : Regions
        Original ROI
    pad : int
        Number of pixels spanning around the original.
    Ishape : (nj, ni)
        shape of image containing all POI, i.e. I.shape
    x, y : 2d array
        global pixel coordinates
    exclusion : 2d array of bool, optional
        If given, needs to match shape of I. Any area that matches this mask
        will be excluded from the rings. The default is None.

    Returns
    -------
    roi_BG : Regions
        New ROI's each forming a ring around the original.
    '''
    if len(args) == 1:
        if type(args[0]) is tuple:
            Ishape = args[0]
            #raise NotImplementedError()
        else:
            x, y = args[0]['x'], args[0]['y']
            Ishape = x.shape
    elif len(args) == 2:
        x, y = args
        Ishape = x.shape
    ny, nx = Ishape
    rslices = []
    rmasks = None if roi.rmasks is None else []
    for k in range(len(roi)):
        # need to expand each ROI to fit a ring
        jslice, islice = roi.rslices[k]
        j0, j1 = jslice.start, jslice.stop
        i0, i1 = islice.start, islice.stop
        # but need to limit expanded ROI so it stays in the frame
        j0_new, j1_new = max(0, j0 - pad), min(ny - 1, j1 + pad)
        i0_new, i1_new = max(0, i0 - pad), min(nx - 1, i1 + pad)
        rslice = (slice(j0_new, j1_new), slice(i0_new, i1_new))
        rslices.append(rslice)
        if roi.rmasks is not None:
            mask0 = roi.rmasks[k]
            mask1 = _np.pad(mask0, pad)
            mask2 = _morph.binary_dilation(mask1)
            for i in range(1, pad):
                mask2 = _morph.binary_dilation(mask2)
            mask3 = mask2 & _np.logical_not(mask1)
            # need to be sure the padding doesn't fall outside the original shape
            jslice_int = slice(j0_new + pad - j0, j1_new - j0 + pad)
            islice_int = slice(i0_new + pad - i0, i1_new - i0 + pad)
            #print(jslice_int, islice_int)
            # if an exclusion mask was passed, need to remove it from mask
            mask4 = mask3[jslice_int, islice_int]
            if exclusion is not None:
                print('scaffold!')
                exclusion_k = exclusion[rslice]
                mask4 = mask4 & _np.logical_not(exclusion_k)
            rmasks.append(mask4)
    return Regions(rslices, rmasks)


def background_fromroiBG(roi_BG, I0):
    bg = _np.zeros(len(roi_BG))
    for k in range(len(roi_BG)):
        I0_k = roi_BG.region_I(k, I0).flatten()
        mask = roi_BG.rmasks[k].flatten()
        bg[k] = _np.median(I0_k[mask])
    return bg



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec 28 18:25:02 2023

@author: chris
"""

import numpy as _np

from . import gen


class DemoData():
    def __init__(self, res=[6000, 5000], tile=300, cr=16, sigma=2, poisson=1):
        self._res = res
        self._tile = tile
        self._sigma = sigma
        self._poisson = poisson
        self._imgen = gen.ChartGen_img(res, cr=cr)

    @property
    def slant(self):
        im, _ = self._imgen.slant_rectangles(self._tile, tile=True, slant=4)
        return gen.image_transfer_simple(
            _np.array(im, dtype='int'),
            sigma=self._sigma, poisson=self._poisson)

    @property
    def cb(self):
        im, _ = self._imgen.checkerboard(self._tile, tile=True)
        return gen.image_transfer_simple(
            _np.array(im, dtype='int'),
            sigma=self._sigma, poisson=self._poisson)

    @property
    def ptgrid(self):
        im, _ = self._imgen.pointgrid(self._tile, tile=True)
        return gen.image_transfer_simple(
            _np.array(im, dtype='int'),
            sigma=self._sigma, poisson=self._poisson)

    '''@property
    def starfield(self):
        im = gen.genstarfield(self._res, nstar, sigma)'''


demoL = DemoData([6000, 5000], 300)
demoM = DemoData([1800, 1500], 300)
demoS = DemoData([600, 500], 150)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 30 08:20:56 2023

@author: chris
"""


import numpy as _np


def _rot_mtrx2d(theta):
    '''Rotation matrix for xy vectors (theta is degrees)'''
    theta = _np.deg2rad(theta)
    A = _np.array([[_np.cos(theta), -_np.sin(theta)],
                  [_np.sin(theta), _np.cos(theta)]])
    return A


def _vert2vect(vertices):
    """For vertices=[(x0, y0), (x1, y1), ...] return 2d vector"""
    vector = _np.array(vertices)
    return vector


def _vect2vert(vector):
    """Return vertices=[(x0, y0), (x1, y1), ...] made from 2d vector"""
    vertices = []
    for vi in vector:
        vertices.append(tuple(vi))
    return vertices


def _offset_vertices(dx, dy, vertices):
    """Apply offsets dx, dy to vertices=[(x0, y0), (x1, y1), ...]"""
    vect = _vert2vect(vertices)
    vect += _np.array([dx, dy])
    vert = _vect2vert(vect)
    return vert


def _rotate_vertices(vertices, theta):
    '''Apply theta(deg) rotation to vertices'''
    vector = _vert2vect(vertices)
    #print(theta)
    A = _rot_mtrx2d(theta)
    #print(A)
    vector = _np.matmul(vector, A)
    vertices = _vect2vert(vector)
    return vertices


def _rotate_2d(xarray, yarray, theta):
    '''Apply theta(deg) rotation to 1d or 2d xarray, yarray'''
    A = _rot_mtrx2d(theta)
    #print(A)
    shape0 = _np.shape(xarray)
    #print('shape0', shape0)
    assert shape0 == _np.shape(yarray), 'shapes disagree'
    # get a vector form of x,y
    vec1 = _np.stack((_np.ravel(xarray), _np.ravel(yarray)), axis=1)
    # rotate by matrix multiplication
    vec2 = _np.matmul(vec1, A)
    # return x, y to original array shape
    x2 = _np.reshape(vec2[:, 0], shape0)
    y2 = _np.reshape(vec2[:, 1], shape0)
    return x2, y2



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar  8 10:07:24 2023

@author: chris
"""

import numpy as _np
from scipy import ndimage as _ndimage


def image_transfer_simple(im0, upsample=None, distortion=None, sigma=None,
                          poisson=None, darklevel=0):
    '''
    Similar to modeling an image transfer from a source, through optics, to a
    sensor, except each step is given by the simplest possible transform.

    Parameters
    ----------
    im0 : 2d array
        input image, normally a flawless generated image
    upsample : float, optional
        Upsample using scipy.ndimage.zoom. The default is None.
    distortion : TYPE, optional
        DESCRIPTION. The default is None.
    sigma : float or [float, float]
        radius of Gaussian blur (see scipy.ndimage.gaussian_filter). The
        default is None.
    poisson : float, optional
        The gain of the sensor, i.e. count/photo-electrons. The default is
        None. If used, poisson statistics create grainy noise as seen on
        a camera.
    darklevel : float, optional
        Add a dark level, in counts

    Returns
    -------
    im : 2d array
        image at sensor after transfer
    '''
    if upsample is None and distortion is None:
        im1 = im0
    else:
        if distortion is None:
            im1 = _ndimage.zoom(im0, upsample, order=0)
        else:
            raise NotImplementedError
    if sigma is None:
        im2 = im1
    else:
        im2 = _ndimage.gaussian_filter(im1, sigma)
    if not _np.isclose(darklevel, 0):
        im2 += darklevel
    if poisson is None:
        im3 = im2
    else:
        gain = poisson
        rgen = _np.random.default_rng()
        im3 = rgen.poisson(im2 / gain) * gain
    return im3


def image_transfer_model(source, sensor=None, optics=None):
    '''
    NOT IMPLEMENTED

    Simulate a model for image transfer from a source, through optics, to a
    sensor.

    Parameters
    ----------
    im0 : TYPE
        DESCRIPTION.
    sensor : TYPE, optional
        DESCRIPTION. The default is None.
    optics : TYPE, optional
        DESCRIPTION. The default is None.
    source : TYPE, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    None.

    '''
    raise NotImplementedError


class ModelSensor():
    def __init__(self):
        raise NotImplementedError


class ModelObject():
    def __init__(self):
        raise NotImplementedError


class ModelImage():
    def __init__(self):
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Feb  1 11:42:43 2025

@author: chris
"""

import numpy as _np
from matplotlib import pyplot as _plt


def plot_Efield(EF, *args):
    '''
    Simple plot of Intensity and Phase of an Efield.
    Returns pyplot figure and subplot objects for further customization.

    Parameters
    ----------
    EF : 2d Array
        Efield
    x, y : 2d Array, optional
        Provide x,y ordinates if you want them plotted instead of
        row,column

    Returns
    -------
    fig : 
        figure
    ax0, ax1 : 
        subplots
    '''
    fig, (ax0, ax1) = _plt.subplots(1, 2)
    I = _np.abs(EF * _np.conjugate(EF))
    phase = _np.arctan2(_np.real(EF), _np.imag(EF))
    # plot intensity
    if len(args) == 2:
        ax0.contourf(*args, I)
        ax0.set_xlabel('X')
        ax0.set_ylabel('Y')
    else:
        ax0.contourf(I)
    ax0.set_title('Intensity')
    # plot phase
    if len(args) == 2:
        ax1.contourf(*args, phase)
        ax1.set_xlabel('X')
        ax1.set_ylabel('Y')
    else:
        ax1.contourf(phase)
    ax1.set_title('Phase')
    return fig, ax0, ax1


def fft_Efield(EFp):
    '''
    Use FFT to compute a Point Spread Function given complex valued Efield.
    See also fft_coords

    Parameters
    ----------
    EFp : 2d array
        Complex valued Efield.

    Returns
    -------
    PSF : 2d array
        Diffracted Point Spread Function
    '''
    EFinf = _np.fft.fftshift(_np.fft.fft2(EFp))
    PSF = _np.abs(EFinf * _np.conjugate(EFinf))
    return PSF


def fftfreq(n, d, wavelength):
    '''
    Warp around np.fft.fftfreq to consider wavelength, and throw in fftshift.
    f = fftshift(0, 1, ..., n/2-1, ..., -1) * wavelength/(d*n) if n is even
    f = fftshift(0, 1, ..., (n-1)/2, ..., -1) * wavelength/(d*n) if n is odd

    Parameters
    ----------
    n : int
        Window length
    d : float
        Sample spacing
    wavelength: float
        Same units as d

    Returns
    -------
    f : 1d array
        Frequency space coordinates of after fft_Efield (radians)

    Example::
        
        # assume EFp has 128 samples at 0.1mm pitch and wavelength is 0.5um;
        f = fftfreq(128, .1, 0.5/1000)
    '''
    f = _np.fft.fftshift(_np.fft.fftfreq(n, d) * wavelength)
    return f


def init_coords(apitch, xpitch, wavelength):
    '''
    Helper to initialize pupil and frequency coordinates while meeting expected
    pitch in each.

    Parameters
    ----------
    apitch : float
        Desired angular pitch in angular (image) space. (radians).
        Negative indicates a min apitch rather than exact (will round down).
    xpitch : float
        Desired linear pitch in pupil space. (same units as wavelength).
        Negative indicates a min xpitch rather than exact (will round down).
        One of xpitch or apitch needs to be negative.
    wavelength : float
        Wavelength of light in units of choice

    Returns
    -------
    apitch : float
    xpitch : float
    npix : int
    '''
    if apitch < 0:
        assert xpitch>0, 'Expected one of apitch or xpitch to be >0'
        pupilspan0 = wavelength / -apitch
        npix = int(_np.ceil(pupilspan0 / xpitch)) + 1
        # after rounding, recalc pupilspan and apitch
        pupilspan = (npix - 1) * xpitch
        apitch = wavelength / pupilspan
    elif xpitch < 0:
        assert apitch>0, 'Expected one of apitch or xpitch to be >0'
        pupilspan = wavelength / apitch
        npix = int(_np.ceil(pupilspan / -xpitch))
        # after rounding, recalc xpitch
        xpitch = pupilspan / npix
    else:
        raise(ValueError, 'Expected one of apitch or xpitch to be <0')
    x_ax = _np.linspace(-pupilspan/2, pupilspan/2, npix)
    a_ax = fftfreq(npix, xpitch, wavelength)
    return apitch, xpitch, npix, x_ax, a_ax


def aperture_Efield(x, y, EF, ODx, ODy, cx=0, cy=0):
    '''
    Zero Efield outside of an elliptical aperture

    Parameters
    ----------
    x, y : 2d array
        Lateral coordinates of EF
    EF : 2d array
        Complex valued Efield
    ODx, ODy : float
        DESCRIPTION.
    cx, cy : float, optional
        x, y centroid location of aperture. The default is 0.

    Returns
    -------
    EF_aprt : 2d array
        Efield after aperture
    '''
    radx = ODx/2
    rady = ODy/2
    x2 = ((x - cx) / radx)**2
    y2 = ((y - cy) / rady)**2
    mask = (x2 + y2) < 1
    EF_aprt = EF * mask
    return EF_aprt


def obscure_Efield(x, y, EF, IDx, IDy, cx=0, cy=0):
    '''
    Zero Efield within an elliptical obscuration

    Parameters
    ----------
    x, y : 2d array
        Lateral coordinates of EF
    EF : 2d array
        Complex valued Efield
    IDx, IDy : float
        DESCRIPTION.
    cx, cy : float, optional
        x, y centroid location of aperture. The default is 0.

    Returns
    -------
    EF_obs : 2d array
        Efield after obscuration
    '''
    radx = IDx/2
    rady = IDy/2
    x2 = ((x - cx) / radx)**2
    y2 = ((y - cy) / rady)**2
    mask = (x2 + y2) >= 1
    EF_obs = EF * mask
    return EF_obs


def illum_2_PSFcam(OD, ID, Fnumber, wavelength, pixel, n=None):
    PSF = 0
    return PSF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Dec 22 21:39:35 2023

@author: chris
"""

import numpy as _np
from skimage.feature import peak_local_max as _peak

from .edge import edgeH_Analysis as _edgeH_Analysis
from .edge import edgeV_Analysis as _edgeV_Analysis
from .edge import esf2mtf as _esf2mtf
from .grad import imageGradients as _imageGrad
from .grad import maskedges_hvx as _mask_ehvx
from .peak import peaksAnalysis as _peaksAnalysis, peaksDownselect as _peaksDownselect
from .roi import Regions as _Regions
from .bg import background_fromroiBG, ringRegions


def unit_norm(a, ref=None):
    '''
    Normalizes a to 0-1.

    unit_norm = (a - max(a, ref)) / (max(a, ref) - min(a, ref))
    '''
    if ref is None:
        mx = _np.max(a)
        mn = _np.min(a)
    else:
        mx = max(_np.max(a), _np.max(ref))
        mn = min(_np.min(a), _np.min(ref))
    return (a - mn) / (mx - mn)


def recipe_checkerboard1(I0, sigma, threshold=0.2):
    '''
    UNFINISHED

    Use the D_hessian to set ROI around saddle points.

    Parameters
    ----------
    I0 : 2d array
        original image
    sigma : float
        sigma for Gaussian blur (pixels)
    threshold : float, optional
        threshold used for mask. The default is 0.2.

    Returns
    -------
    imG : dict of 2d arrays
        See imageGradients, also will have 'mask'
    roi : Regions object
        Regions of Interest
    peaks : dict of 1d arrays
    '''
    imG = _imageGrad(I0, sigma, ['D_hessian'])
    D_h = imG['D_hessian']
    mask = -D_h / _np.max(-D_h) > threshold
    imG['mask'] = mask
    imG['f_saddle'] = -D_h
    #poi = _peak(-D_h, min_distance=sigma)
    #roi = _Regions.from_poi(poi, sigma*2, sigma*2, imG)
    roi = _Regions.from_mask(mask, imG)
    peaks = _peaksAnalysis(roi, imG, 'f_saddle')
    return imG, roi, peaks


def recipe_checkerboard2(I0, sigma, threshold=0.1):
    '''
    UNFINISHED

    Use D_hessian combined with I_r to find saddle points. This is more
    advanced than D_hessian alone, it rejects corner features where the slope
    magnitude is not 0.

    Parameters
    ----------
    I0 : 2d array
        original image
    sigma : float
        sigma for Gaussian blur (pixels)
    threshold : float, optional
        threshold of mask. The default is 0.1.

    Returns
    -------
    imG : dict of 2d arrays
        See imageGradients, also will have 'mask'
    roi : Regions object
        Regions of Interest
    peaks : dict of 1d arrays
    '''
    imG = _imageGrad(I0, sigma, ['D_hessian', 'I_r'])
    # upside-down, 0-1 norm, I_r (slope magnitude) makes a filter function
    # that will reject corners that aren't actual saddle-points
    f_minslope = unit_norm(-imG['I_r'])
    D_h = imG['D_hessian']
    # f_saddle is max where D_hessian is min and where I_r is min
    f_saddle = _np.maximum(0, -D_h * f_minslope)
    mask = f_saddle / _np.max(f_saddle) > threshold
    imG['mask'] = mask
    imG['f_saddle'] = f_saddle
    poi = _peak(f_saddle * mask, min_distance=sigma)
    roi = _Regions.from_poi(poi, sigma*2, sigma*2, imG)
    peaks = _peaksAnalysis(roi, imG, 'f_saddle')
    return imG, roi, peaks


def recipe_slantedge(I0, sigma, threshold=0.1, min_area=20):
    '''
    UNFINISHED

    Slantedge analysis of PSF and MTF.

    Parameters
    ----------
    I0 : 2d array
        original image
    sigma : float
        sigma for Gaussian blur (pixels)
    threshold : float, optional
        threshold of mask. The default is 0.1.

    Returns
    -------
    imG : TYPE
        DESCRIPTION.
    roi_v_pos : Regions
        DESCRIPTION.
    roi_v_neg : Regions
        DESCRIPTION.
    roi_h_pos : Regions
        DESCRIPTION.
    roi_h_neg : Regions
        DESCRIPTION.
    '''
    imG = _imageGrad(I0, sigma, ['I_x', 'I_y'])
    _mask_ehvx(imG, threshold)
    roi_vp = _Regions.from_mask(imG['mask_v_pos'], imG, min_area=min_area)
    roi_vn = _Regions.from_mask(imG['mask_v_neg'], imG, min_area=min_area)
    roi_hp = _Regions.from_mask(imG['mask_h_pos'], imG, min_area=min_area)
    roi_hn = _Regions.from_mask(imG['mask_h_neg'], imG, min_area=min_area)
    rval = dict(imG=imG, roi_vp=roi_vp, roi_vn=roi_vn, roi_hp=roi_hp,
                roi_hn=roi_hn)
    esf_vp = []
    rval['esf_vp'] = esf_vp
    for k in range(len(roi_vp.rslices)):
        imG_k = roi_vp.region_imG(k, imG)
        ESF, dx, details = _edgeV_Analysis(imG_k)
        LSF, MTF = _esf2mtf(ESF, dx)
        esf_vp.append((k, ESF, dx, LSF, MTF, details))
    esf_vn = []
    rval['esf_vn'] = esf_vn
    for k in range(len(roi_vn.rslices)):
        imG_k = roi_vn.region_imG(k, imG)
        ESF, dx, details = _edgeV_Analysis(imG_k)
        LSF, MTF = _esf2mtf(ESF, dx)
        esf_vn.append((k, ESF, dx, LSF, MTF, details))
    esf_hp = []
    rval['esf_hp'] = esf_hp
    for k in range(len(roi_hp.rslices)):
        imG_k = roi_hp.region_imG(k, imG)
        ESF, dy, details = _edgeH_Analysis(imG_k)
        LSF, MTF = _esf2mtf(ESF, dy)
        esf_hp.append((k, ESF, dy, LSF, MTF, details))
    esf_hn =[]
    rval['esf_hn'] = esf_hn
    for k in range(len(roi_hn.rslices)):
        imG_k = roi_hn.region_imG(k, imG)
        ESF, dy, details = _edgeH_Analysis(imG_k)
        LSF, MTF = _esf2mtf(ESF, dy)
        esf_hn.append((k, ESF, dy, LSF, MTF, details))
    return rval


def recipe_star1(I0, sigma, threshold=0.1):
    '''
    Find stars (or beams) in an image in a single pass.
    Using curvature and slope-magnitude to find ROI.

    Parameters
    ----------
    I0 : 2d array
        original image
    sigma : float
        sigma for Gaussian blur (pixels)
    threshold : float, optional
        threshold of mask. The default is 0.1.

    Returns
    -------
    imG : dict of 2d arrays
        See imageGradients
    roi : Regions object
        Regions of Interest
    peaks : dict of 1d arrays
    '''
    imG = _imageGrad(I0, sigma, ['curve', 'I_r'])
    maskA = imG['I_r']/_np.max(imG['I_r']) > threshold
    maskB = imG['curve']/_np.min(imG['curve']) > threshold
    # 1st iteration:
    # peak mask = (slope mag is rel. max) | (curve is rel. min)
    mask = maskA | maskB
    roi = _Regions.from_mask(mask, imG)
    peaks = _peaksAnalysis(roi, imG, 'I0', ['xsigma', 'ysigma'])
    return imG, roi, peaks


def recipe_star2(I0, sigma, threshold=0.1, nsigma=4, E_amin=1, E_rmin=.02):
    '''
    Find stars (or beams) in an image with a more noise resistant process:
        * First pass; find ROI using curvature and slope-magnitude
        * Downselect ROI w (E < E_amin) or (E/mx(E) < E_rmin) or cx,xsigma,etc=nan
        * Surround every ROI with a ring to get background
        * Calculate peak parameters
        * Use (xsigma, ysigma) * nsigma to set revised ROI
        * Ring based background again
        * Recalc peak parameters

    Parameters
    ----------
    I0 : 2d array
        original image
    sigma : float
        sigma for Gaussian blur (pixels)
    threshold : float, optional
        threshold for first pass mask on curvature and slope-magnitude.
        The default is 0.1.
    nsigma : float, optional
        ROI radius in nsigma, defines the size of the second-pass ROI.
        The default is 4.

    Returns
    -------
    imG : dict of 2d arrays
        See imageGradients
    roi : Regions object
        Regions of Interest
    peaks : dict of 1d arrays
        For each peak, 
    xdata : dict
        Various extra results, only for curiosity
    '''
    imG = _imageGrad(I0, sigma, ['curve', 'I_r'])
    maskA = imG['I_r']/_np.max(imG['I_r']) > threshold
    maskB = imG['curve']/_np.min(imG['curve']) > threshold
    # 1st iteration:
    # peak mask = (slope mag is rel. max) | (curve is rel. min)
    mask = maskA | maskB
    #imG['mask'] = mask
    roiA = _Regions.from_mask(mask, imG)
    roibgA = ringRegions(roiA, 7, I0.shape)
    bgA = background_fromroiBG(roibgA, I0)
    peaksA = _peaksAnalysis(roiA, imG, 'I0', ['xsigma', 'ysigma'], bg=bgA)
    # downselect peaks and roi
    mxE = _np.max(peaksA['E'])
    isvalid = ((peaksA['E'] > E_amin) & (peaksA['E']/mxE > E_rmin)
               & ~_np.isnan(peaksA['cx']) & ~_np.isnan(peaksA['cy'])
               & ~_np.isnan(peaksA['xsigma']) & ~_np.isnan(peaksA['ysigma']))
    peaksB = _peaksDownselect(peaksA, isvalid)
    # 2nd iteration:
    # use xsigma, ysigma from prior to revise ROI
    #poiA = list(zip(peaksA['cy'], peaksA['cx']))
    poiB = _np.stack((peaksB['cy'], peaksB['cx']), axis=1)
    #print(poiA)
    roi = _Regions.from_poi(poiB, nsigma*peaksB['xsigma'], nsigma*peaksB['ysigma'], imG)
    roibg = ringRegions(roi, 7, I0.shape)
    bg = background_fromroiBG(roibg, I0)
    peaks = _peaksAnalysis(roi, imG, 'I0', ['xsigma', 'ysigma'], bg=bg)
    xdata = dict(roibgA=roibgA, peaksA=peaksA, bgA=bgA, roibg=roibg, bg=bg)
    return imG, roi, peaks, xdata




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jan 20 09:37:43 2024

@author: chris
"""

from matplotlib import pyplot as plt
import numpy as _np
from scipy import interpolate as _interp


def edgeV_Analysis(imG_k, upsample=1, cutoffs=(85, .85, 2)):
    '''
    Parameters
    ----------
    imG_k : dict of 2d arrays
        Contains 'I0', 'I_y', 'ones'. Has been trimmed to an ROI.
    upsample : float, optional.
        Increase sampling by this factor, e.g. 2x. The default is 1x.
    cutoffs : (float, float, int), optional
        (percentile0-100, threshold0-1, min_offset).
        The default is (85, .85, 2).
        Cutoffs are used to select rows that have enough good samples.
        e.g. (100, .5, 0) means FWHM (of samples per row).
        e.g. (50, 1, 0) means >median (of samples per row).

    Returns
    -------
    ESF : 1d array
        Edge Spread Function
    dx : 1d array
        Sample column w.r.t. center of edge. Spacing is 1/upsample
    edgeV_detail : TYPE
        dict of various intermediates.

    Pseudo-code:
        1. sum(I_x), cx, n for each row
        2. mask: select rows above a cutoff criterea for n
        3. line fit cx[mask] vs y[mask]
        4. dx = x - cx
        5. flatten I0 vs dx
        6. resample on new x
    '''
    I_x = imG_k['I_x']
    ones = imG_k['ones']
    nx = I_x.shape[1]
    x_ax = _np.arange(nx)
    # stats per row based on I_x
    n = ones.sum(1)
    Ix = I_x.sum(1)
    cx_raw = (I_x*x_ax).sum(1) / Ix
    # mask off rows that don't have enough samples
    # maybe think this out more later
    percentile, threshold, min_offset = cutoffs
    high = _np.percentile(n, percentile)
    mask = n > min((high - min_offset), (threshold * high))
    good, = _np.nonzero(mask)
    # fit y, cx to a line, ignoring masked off data
    y_ax = _np.arange(I_x.shape[0])
    y2 = y_ax[mask]
    cx2 = cx_raw[mask]
    c, _, _, _ = _np.linalg.lstsq(_np.vstack([y2, _np.ones(len(y2))]).T,
                                  cx2, rcond=-1)
    m = c[0]
    b = c[1]
    cx_fit = b + m*y_ax
    # need 2d arrays: ones; mask, I0, dx; x position w.r.t. the edge
    x, y = _np.meshgrid(x_ax, y_ax)
    dx = x - cx_fit.reshape(-1, 1)
    ones = _np.array(ones * mask.reshape(-1, 1), dtype=bool)
    I0 = imG_k['I0']
    # resample irregular (I0 vs dx) onto a new grid
    ones_flat = ones.flatten()
    dx_flat = dx.flatten()[ones_flat]
    y_flat = y.flatten()[ones_flat]
    I0_flat = I0.flatten()[ones_flat]
    nx2 = int(_np.median(n[mask]))
    mnmx = (nx2-1/upsample)/2
    dx2, y2 = _np.meshgrid(_np.linspace(-mnmx, mnmx, nx2*upsample),
                           good)
    #dx2, y2 = _np.meshgrid(_np.arange(-(nx2-1)/2, nx2-(nx2-1)/2),
    #                       good)
    ESF2 = _interp.griddata((dx_flat, y_flat), I0_flat, (dx2, y2), 'cubic',
                            fill_value=0)
    ones2 = _interp.griddata((dx.flatten(), y.flatten()), ones.flatten(),
                             (dx2, y2), 'cubic', fill_value=0)
    #rbf = _interp.Rbf(dx_flat, y_flat, I0_flat)
    #ESF2 = rbf(dx2, y2)
    #rbf = _interp.Rbf(dx.flatten(), y.flatten(), ones.flatten())
    #ones2 = rbf(dx2, y2)
    # project onto dx axis and take average
    nc = ones2.sum(0)
    ESF = (ESF2*ones2).sum(0) / nc
    # intermediate details
    edgeV_detail = dict(cx_raw=cx_raw, cx_fit=cx_fit, Ix=Ix, n=n, mask=mask,
                        y_ax=y_ax, dx=dx, I0=I0*ones, ones=ones,
                        dx_flat=dx_flat, I0_flat=I0_flat, dx2=dx2, ESF2=ESF2,
                        ones2=ones2)
    return ESF, dx2[0, :], edgeV_detail


def plot_edgeV_data(ESF, dx, edgeV_details):
    plt.figure()
    plt.subplot(1, 2, 1)
    plt.imshow(edgeV_details['I0'], cmap='gray')
    mask = edgeV_details['mask']
    cx_raw = edgeV_details['cx_raw'][mask]
    cx_fit = edgeV_details['cx_fit'][mask]
    y_ax = edgeV_details['y_ax'][mask]
    plt.plot(cx_raw, y_ax, 'r+')
    plt.plot(cx_fit, y_ax)
    plt.subplot(1, 2, 2)
    dx_flat = edgeV_details['dx_flat']
    I0_flat = edgeV_details['I0_flat']
    plt.plot(dx_flat, I0_flat, '.')
    plt.plot(dx, ESF)


def edgeH_Analysis(imG_k, upsample=1, cutoffs=(85, .85, 2)):
    '''
    Parameters
    ----------
    imG_k : dict of 2d arrays
        Contains 'I0', 'I_x', 'ones'.
    upsample : float, optional.
        Increase sampling by this factor, e.g. 2x. The default is 1x.
    cutoffs : (float, float, int), optional
        (percentile0-100, threshold0-1, min_offset).
        The default is (85, .85, 2).
        Cutoffs are used to select column that have enough good samples.
        e.g. (100, .5, 0) means FWHM (of samples per column).
        e.g. (50, 1, 0) means >median (of samples per column).

    Returns
    -------
    ESF : 1d array
        Edge Spread Function
    dy : 1d array
        Sample column w.r.t. center of edge. Spacing is 1/upsample
    edgeV_detail : dict
        dict of various intermediates.

    Pseudo-code:
        1. sum(I_y), cy, n for each row
        2. mask: select rows above a cutoff criterea for n
        3. line fit cy[mask] vs x[mask]
        4. dx = x - cx
        5. flatten I0 vs dy
        6. resample on new y
    '''
    I_y = imG_k['I_y']
    ones = imG_k['ones']
    ny = I_y.shape[0]
    y_ax = _np.arange(ny)
    # stats per row based on I_y
    n = ones.sum(0)
    Iy = I_y.sum(0)
    cy_raw = (I_y*y_ax.reshape(-1, 1)).sum(0) / Iy
    # mask off rows that don't have enough samples
    # maybe think this out more later
    percentile, threshold, min_offset = cutoffs
    high = _np.percentile(n, percentile)
    mask = n > min((high - min_offset), (threshold * high))
    good, = _np.nonzero(mask)
    # fit y, cx to a line, ignoring masked off data
    x_ax = _np.arange(I_y.shape[1])
    x2 = x_ax[mask]
    cy2 = cy_raw[mask]
    c, _, _, _ = _np.linalg.lstsq(_np.vstack([x2, _np.ones(len(cy2))]).T,
                                  cy2, rcond=-1)
    m = c[0]
    b = c[1]
    cy_fit = b + m*x_ax
    # need 2d arrays: ones; mask, I0, dy; y position w.r.t. the edge
    x, y = _np.meshgrid(x_ax, y_ax)
    dy = y - cy_fit
    ones = _np.array(ones * mask, dtype=bool)
    I0 = imG_k['I0']
    # resample irregular (I0 vs dx) onto a new grid
    ones_flat = ones.flatten()
    dy_flat = dy.flatten()[ones_flat]
    x_flat = x.flatten()[ones_flat]
    I0_flat = I0.flatten()[ones_flat]
    ny2 = int(_np.median(n[mask]))
    mnmx = (ny2-1/upsample)/2
    x2, dy2 = _np.meshgrid(good,
                           _np.linspace(-mnmx, mnmx, ny2*upsample))
    #dx2, y2 = _np.meshgrid(_np.arange(-(nx2-1)/2, nx2-(nx2-1)/2),
    #                       good)
    ESF2 = _interp.griddata((x_flat, dy_flat), I0_flat, (x2, dy2), 'cubic',
                            fill_value=0)
    ones2 = _interp.griddata((x.flatten(), dy.flatten()), ones.flatten(),
                             (x2, dy2), 'cubic', fill_value=0)
    #rbf = _interp.Rbf(dx_flat, y_flat, I0_flat)
    #ESF2 = rbf(dx2, y2)
    #rbf = _interp.Rbf(dx.flatten(), y.flatten(), ones.flatten())
    #ones2 = rbf(dx2, y2)
    # project onto dx axis and take average
    nc = ones2.sum(1)
    ESF = (ESF2*ones2).sum(1) / nc
    # intermediate details
    edgeV_detail = dict(cy_raw=cy_raw, cy_fit=cy_fit, Iy=Iy, n=n, mask=mask,
                        x_ax=x_ax, dyx=dy, I0=I0*ones, ones=ones,
                        dy_flat=dy_flat, I0_flat=I0_flat, dy2=dy2, ESF2=ESF2,
                        ones2=ones2)
    return ESF, dy2[:, 0], edgeV_detail


def plot_edgeH_data(ESF, dy, edgeV_details):
    plt.figure()
    plt.subplot(2, 1, 1)
    plt.imshow(edgeV_details['I0'], cmap='gray')
    mask = edgeV_details['mask']
    cy_raw = edgeV_details['cy_raw'][mask]
    cy_fit = edgeV_details['cy_fit'][mask]
    x_ax = edgeV_details['x_ax'][mask]
    plt.plot(x_ax, cy_raw, 'r+')
    plt.plot(x_ax, cy_fit)
    plt.subplot(2, 1, 2)
    dy_flat = edgeV_details['dy_flat']
    I0_flat = edgeV_details['I0_flat']
    plt.plot(dy_flat, I0_flat, '.')
    plt.plot(dy, ESF)


def esf2mtf(ESF, dx=None):
    '''
    Parameters
    ----------
    ESF : 1d array
        Edge-Spread-Function of length n
    dx : 1d array, optional
        x coordinates.

    Returns
    -------
    LSF : 1d array
        Line-Spread-Function, of length n-1, normalized so integral is 1.
    MTF : 1d-array
        Modulation-Transfer-Function = abs(fft(LSF))[:(n-1)//2]
    '''
    pkpk = _np.max(ESF) - _np.min(ESF)
    #delta = dx[1] - dx[0]
    #delta = _np.median(_np.diff(dx))
    LSF = _np.diff(ESF/pkpk)
    n = len(LSF)
    MTF = _np.abs(_np.fft.fft(LSF))[:n//2]
    return LSF, MTF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 24 14:56:18 2023

@author: chris
"""

from matplotlib import pyplot as _plt
import numpy as _np
from scipy import optimize as _opt
import warnings

from .math import _rotate_2d


def eval_at_POI(a, POI):
    '''
    Parameters
    ----------
    a : 2d array
        Possibly image used in the peak finder, or any other 2d array to eval
        at peaks.
    POI : array([[j0, i0], [j1, i1], ...])
        Points-of-Interest (same as peaks from
        skimage.feature.peak_local_max)

    Returns
    -------
    ev : array
        evaluation of image at each point
    '''
    # flatten the peak indices
    i = POI[:, 0]
    j = POI[:, 1]
    idx = _np.ravel_multi_index((i, j), a.shape)
    ev = a.flatten()[idx]
    return ev


def find_major_sigma(x, y, Inorm):
    '''
    Uses second moments and rotation matrices to find the major axis of an
    beam or peak image.

    Parameters
    ----------
    x, y : 2d arrays
        x, y with respect to the centroid
    Inorm : 2d array
        normalized image

    Returns
    -------
    maj_sigma, min_sigma : float
        sigma (second-moment) along minor, major axes
    theta : float
        rotation angle (degrees) from y axis to the major axis. I.e.
        if you rotate x,y,I by -theta then you'll see the major axis along
        y, and minor axis along x.
    '''
    #print('xshape %s yshape %s' % (x.shape, y.shape))
    def _rotated_xysigma(theta):
        x2, y2 = _rotate_2d(x, y, theta)
        xsigma = _np.sqrt(_np.sum(x2**2 * Inorm))
        ysigma = _np.sqrt(_np.sum(y2**2 * Inorm))
        return xsigma, ysigma
    def _rotation_merit(x):
        theta = x[0]
        #print('rotating', theta)
        xsigma, ysigma = _rotated_xysigma(theta)
        #print('xsigma: %0.1f' % xsigma)
        return xsigma
    res =_opt.minimize(_rotation_merit, [0], bounds=[(-180, 180)])
    #assert res.success, ''
    theta = res.x[0]
    min_sigma, maj_sigma = _rotated_xysigma(theta)
    return maj_sigma, min_sigma, -theta


def peaksAnalysis(roi, imG, key, require=['cx', 'cy'], bg=None):
    '''
    For each region in roi, analyze properties of a peak in that region.

    Parameters
    ----------
    roi : Regions
        Regions of interest
    imG : dict of 2d arrays
        Normally, the original image array I0 and its gradients. Needs to have
        'x', 'y', and key at least.
    key : str
        The image in imG to be analyzed
    require : list of str, optional
        Specify min set of analyses that will run. The default is ['cx', 'cy'].
        Dependencies left unspecified will be auto resolved. For example,
        'xsigma' alone will imply 'cx'.

    Returns
    -------
    peaks : dict of arrays
        Table of results.

    Names of Arrays
    ---------------
    | E : integral
    | cx, cy : fractional centroids in global coordinates
    | xsigma, ysigma : sigma is a second moment, oriented to x, y axes.
    '''
    peaks = {}
    do_xysigma = 'xsigma' in require or 'ysigma' in require
    do_majsigma = 'maj_sigma' in require or 'min_sigma' in require
    do_cent = do_xysigma or do_majsigma or ('cx' in require or 'cy' in require)
    do_E = do_cent or 'E' in require
    if do_E:
        E = _np.zeros(len(roi))
    if do_cent:
        cx, cy = _np.zeros(len(roi)), _np.zeros(len(roi))
    if do_xysigma:
        xsigma, ysigma = _np.zeros(len(roi)), _np.zeros(len(roi))
    if do_majsigma:
        maj_sigma = _np.zeros(len(roi))
        min_sigma = _np.zeros(len(roi))
        theta = _np.zeros(len(roi))
    for k in range(len(roi)):
        imG_k = roi.region_imG(k, imG, keys=['x', 'y', key])
        xk = imG_k['x']
        yk = imG_k['y']
        I = imG_k[key] if bg is None else _np.maximum(0, imG_k[key] - bg[k])
        if do_E:
            E[k] = _np.sum(I)
        if do_cent:
            cy[k] = _np.sum(yk * I) / E[k]
            cx[k] = _np.sum(xk * I) / E[k]
        if do_xysigma:
            xsigma[k] = _np.sqrt(_np.sum(
                (xk - cx[k])**2 * I) / E[k])
            ysigma[k] = _np.sqrt(_np.sum(
                (yk - cy[k])**2 * I) / E[k])
        if do_majsigma:
            maj_sigma[k], min_sigma[k], theta[k] = find_major_sigma(
                xk - cx[k], yk - cy[k], I/E[k])
    if do_E:
        peaks['E'] = E
    if do_cent:
        peaks['cx'] = cx
        peaks['cy'] = cy
    if do_xysigma:
        peaks['xsigma'] = xsigma
        peaks['ysigma'] = ysigma
    if do_majsigma:
        peaks['maj_sigma'] = maj_sigma
        peaks['min_sigma'] = min_sigma
        peaks['theta'] = theta
    return peaks


def peaksPlotter(roi, I0, peaks, tag='k+', cmap='jet', header=None):
    _, ax = _plt.subplots(1, 1, layout='tight')
    ax.imshow(I0, cmap=cmap)
    if header is None:
        header = list(peaks.keys())
    Npeaks = len(peaks[header[0]])
    #print('Npeaks', Npeaks)
    for i in range(Npeaks):
        cx, cy = peaks['cx'][i], peaks['cy'][i]
        key = header[0]
        text = '%s %0.3e' % (key, peaks[key][i])
        for key in header[1:]:
            text += '\n%s %0.3e' % (key, peaks[key][i])
        #print(text)
        ax.plot(cx, cy, tag)
        ax.text(cx+10, cy, text,
                bbox=dict(boxstyle='round', facecolor='wheat', alpha=.6))
    return ax


def peaksDownselect(peaks, isvalid):
    '''
    Parameters
    ----------
    peaks : dict of uniform arrays
        Typical form of peaks is e.g. {'E':[E0, E1, ...], 'cx':..}
        Where all arrays are the same length.
    isvalid : array of bool
        Length needs to match the length of arrays in peaks.

    Returns
    -------
    rpeaks : dict of uniform (shorter) arrays
        Downselect all the peaks that aren't valid.
    '''
    rpeaks = {}
    for key, val in peaks.items():
        rpeaks[key] = val[isvalid]
    return rpeaks


class PeaksAnalysis():
    '''
    DEPRECIATED

    Properties
    ----------
    energy : array
        volume (energy) under each peak.
    centroid : arrays ([cy], [cx])
        centroid coordinates of each peak. cy means fractional row,
        cx means fractional column
    xy_sigma : arrays ([xsigma], [ysigma])
        Sigma is a second moment. xsigma, ysigma are oriented to x, y axes.
    major_sigma : arrays ([maj_sigma], [min_sigma], [theta])
        Sigma is a second moment. maj_sigma, min_sigma are major, minor sigmas
        which are oriented to theta
    d4sigma : d4s_x, d4s_y, d4s_maj, d4s_min, theta
        Actually it's just the same as sigma, but 4x, i.e. 4*(second-moment).
        And that makes it a well used measurement of diameter.
    '''
    def __init__(self, roi, imG, key):
        '''
        Parameters
        ----------
        roi : Regions object
            defines Region of Interest for each peak
        imG : dict of 2d arrays
            must contain key and 'x', 'y' at a minimum
        '''
        warnings.warn("Use peaksAnalysis function instead of PeaksAnalysis class")
        self.imG = imG
        self.key = key
        self.N = len(roi)
        self.roi = roi

    @property
    def energy(self):
        if not hasattr(self, '_energy'):
            energy = _np.zeros(self.N)
            for k in range(self.N):
                
                energy[k] = _np.sum(self.roi.region_I(k, self.I))
            self._energy = energy
        return self._energy

    @property
    def centroid(self):
        if not hasattr(self, '_centroid'):
            norm = self.energy
            cy, cx = _np.zeros(self.N), _np.zeros(self.N)
            for k in range(self.N):
                # FIXME
                xframe = self.roi.xframes[k]
                yframe = self.roi.yframes[k]
                cy[k] = _np.sum(yframe * self._light[k]) / norm[k]
                cx[k] = _np.sum(xframe * self._light[k]) / norm[k]
            self._centroid = cy, cx
        return self._centroid

    @property
    def xy_sigma(self):
        if not hasattr(self, '_xy_sigma'):
            cy, cx = self.centroid
            xframes = self.ROI.xframes
            yframes = self.ROI.yframes
            norm = self.energy
            xsigma = _np.zeros(self.Npoi)
            ysigma = _np.zeros(self.Npoi)
            for i in range(self.Npoi):
                xsigma[i] = _np.sqrt(_np.sum(
                    (xframes[i] - cx[i])**2 * self._light[i] / norm[i]))
                ysigma[i] = _np.sqrt(_np.sum(
                    (yframes[i] - cy[i])**2 * self._light[i] / norm[i]))
            self._xy_sigma = xsigma, ysigma
        return self._xy_sigma

    @property
    def major_sigma(self, i):
        if not hasattr(self, '_major_sigma'):
            cy, cx = self.centroid
            xframes = self.ROI.xframes
            yframes = self.ROI.yframes
            norm = self.energy
            maj_sigma = _np.zeros(self.Npoi)
            min_sigma = _np.zeros(self.Npoi)
            thetas = _np.zeros(self.Npoi)
            for i in range(self.Npoi):
                x = xframes[i] - cx[i]
                y = yframes[i] - cy[i]
                maj_sigma, min_sigma, theta = find_major_sigma(
                    x, y, self._light[i]/norm[i])
                maj_sigma[i] = maj_sigma
                min_sigma[i] = min_sigma
                thetas[i] = theta
            self._major_sigma = maj_sigma, min_sigma, thetas
        return self._major_sigma

    def encircled(self, level):
        raise NotImplementedError

    def iteratePeaks_xysigma2ROI(self):
        '''
        NOT IMPLEMENTED

        Use self.xy_sigma to get new ROI and return an iteration of self.
        '''
        raise NotImplementedError

    def plot_singlepeak(self, k):
        # FIXME
        fig = _plt.figure()
        ax1 = fig.add_subplot(1, 2, 1)
        e = self.energy
        cy, cx = self.centroid
        dx = self.ROI.dx
        dy = self.ROI.dy
        ax1.imshow(self._light[k], cmap='gray')
        ax1.plot(cx[k] - dx[k], cy[k] - dy[k], '.r')
        msg = 'E %0.2e at (%0.2f, %0.2f)' % (e[k], cx[k], cy[k])
        _plt.title(msg)
        ax2 = fig.add_subplot(1, 2, 2)
        ax2.imshow(self.ROI.masks[k], cmap='gray')
        _plt.title('Mask')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Feb 25 14:49:13 2023

@author: chris
"""

import numpy as _np

from .chart import itr_checkerboard, itr_slant_rectangles, _naming_grid, limit_polygon2frame

# try imports for extras_require
try:
    import gdstk as _gdstk
    _GDS_isavailable = True
except ImportError:
    _GDS_isavailable = False
try:
    from reportlab.pdfgen.canvas import Canvas as _Canvas
    from reportlab.lib import units as _units
    from reportlab.lib import pagesizes as _ps
    from reportlab.graphics.shapes import Polygon, Drawing, Rect
    from reportlab.lib.colors import Color as _Color
    _PDF_isavailable = True
except ImportError:
    _PDF_isavailable = False


if _GDS_isavailable:
    class ChartGen_gds():
        def __init__(self):
            raise NotImplementedError
else:
    class ChartGen_gds():
        def __init__(*args, **kwargs):
            raise ImportError('missing optional gdstk, try "pip install opticiq[extras]"')


if _PDF_isavailable:
    _P_IND = dict(
        letter=_ps.letter, legal=_ps.legal, TABLOID=_ps.TABLOID,
        tabloid=_ps.TABLOID, A0=_ps.A0, A1=_ps.A1, A2=_ps.A2, A3=_ps.A3, A4=_ps.A4,
        A5=_ps.A5, A6=_ps.A6, A7=_ps.A7, A8=_ps.A8, A9=_ps.A9, B0=_ps.B0, B1=_ps.B1,
        B2=_ps.B2, B3=_ps.B3, B4=_ps.B4, B5=_ps.B5, B6=_ps.B6, B7=_ps.B7, B8=_ps.B8,
        B9=_ps.B9, B10=_ps.B10, C0=_ps.C0, C1=_ps.C1, C2=_ps.C2, C3=_ps.C3, C4=_ps.C4,
        C5=_ps.C5, C6=_ps.C6, C7=_ps.C7, C8=_ps.C8, C9=_ps.C9, C10=_ps.C10)
    _UNIT_IND = dict(
        inch=_units.inch, mm=_units.mm, cm=_units.cm)

    class ChartGen_pdf():
        '''
        Vector pdf chart generators for optical testing.

        Example::
            
            g = ChartGen_pdf(6, 'mm', 'A4')
            c = g.checkerboard(10, tile=True)  # 10x10mm checkers
            c.save()
        '''
        def __init__(self, border, unit, pagesize, cr=None, light=1, invert=False):
            '''
            Parameters
            ----------
            border : float
                size of the page border, in unit
            unit : str
                one of 'inch', 'mm', 'cm'
            pagesize : str
                one of 'letter', 'legal', 'tabloid', 'A0', 'A1', 'A2', 'A3',
                'A4', 'A5', 'A6', 'A7', 'A8', 'A9', 'B0', 'B1', 'B2', 'B3',
                'B4', 'B5', 'B6', 'B7', 'B8', 'B9', 'B10', 'C0', 'C1', 'C2',
                'C3', 'C4', 'C5', 'C6', 'C7', 'C8', 'C9', 'C10'
            cr : float >1, optional
                Contrast ratio, with None meaning inf. The default is None.
            light : float 0-1, optional
            invert : bool, optional
                Inverts the pattern (white on black instead of black on white).
                The default is False, because paper is nominally white and
                paper is probably more uniform than ink or toner.
            '''
            self._basename = pagesize + '_'
            dark = 0 if cr is None else light/(1 + cr)
            if invert:
                self._fill = _Color(light, light, light)
                self._bg = _Color(dark, dark, dark)
            else:
                self._fill = _Color(dark, dark, dark)
                self._bg = None if _np.isclose(light, 1) else _Color(light, light, light)
            self._unit = unit
            self._pagesize = pagesize
            unit = _UNIT_IND[unit]
            pagesize = _P_IND[pagesize]
            self._border = border*unit
            width = pagesize[0] - 2*border*unit
            height = pagesize[1] - 2*border*unit
            assert (width > 0 and height > 0), 'Borders exceeded canvas size'
            self.frame = [width, height]

        def _get_canvas(self, fname):
            '''
            Returns new blank canvas and drawing
            '''
            pagesize = _P_IND[self._pagesize]
            c = _Canvas(fname, pagesize=pagesize)
            if self._bg is not None:
                #r = Rect(0, 0, self.frame[0], self.frame[1])
                #d.setProperties()
                c.setFillColor(self._bg)
                c.setStrokeColor(self._bg)
                b = self._border
                c.rect(b, b, self.frame[0], self.frame[1], fill=1)
            c.setFillColor(self._fill)
            c.setStrokeColor(self._fill)
            d = Drawing(*tuple(self.frame))
            return c, d

        def checkerboard(self, p, tile=True):
            '''
            Parameters
            ----------
            p : float or [float, float]
                (tile) - interval between patterns in unit,
                or (stretch) - number of patterns
            tile : bool, optional
                (True) - tile at even interval ; (False)- stretch pattern to fit the frame

            Returns
            -------
            canvas : reporltab Canvas
                test chart; checkerboard
            '''
            unit = _UNIT_IND[self._unit]
            if tile:
                p = _np.array(p) * unit
            b = self._border
            naming = _naming_grid(self.frame, p, tile=tile, unit=unit)
            naming += self._unit
            fname = self._basename + 'checker' + naming
            c, d = self._get_canvas(fname)
            for x1, y1, x2, y2 in itr_checkerboard(self.frame, p, tile=tile):
                r = Rect(x1, y1, x2 - x1, y2 - y1)
                #c.rect(x1 + b, y1 + b, x2 - x1, y2 - y1, fill=1, stroke=0)
                r.fillColor = self._fill
                r.strokeWidth = 0
                r.strokeColor = self._fill
                d.add(r)
            d.drawOn(c, b, b)
            return c

        def slant_rectangles(self, p, tile=True, slant=5, fill_factor=0.5):
            '''
            Parameters
            ----------
            p : float or [float, float]
                (tile) - interval between patterns in unit,
                or (stretch) - number of patterns
            tile : bool, optional
                (True) - tile at even interval ; (False) - stretch pattern to fit the frame
            slant : float, optional
                angle in degrees to slant the squares (default = 5 degrees)
            fill_factor : float 0 to ~.95 (default = 0.5)
                low fill factor means more space between slanted squares

            Returns
            -------
            canvas : reporltab Canvas
                test chart; checkerboard
            '''
            unit = _UNIT_IND[self._unit]
            if tile:
                p = _np.array(p) * unit
            naming = _naming_grid(self.frame, p, tile=tile, unit=unit)
            naming += self._unit
            fname = (self._basename + 'slantrect' + naming
                     + '_fill%0.2fslant%0.1fdeg' % (fill_factor, slant))
            c, d = self._get_canvas(fname)
            for vertices in itr_slant_rectangles(
                    self.frame, p, tile=tile, slant=slant, fill_factor=fill_factor):
                vertices = limit_polygon2frame(self.frame, vertices)
                # vertices are a zipped list, need to flatten it for Polygon
                vertices = [el for xy in vertices for el in xy]
                pg = Polygon(vertices)
                pg.fillColor = self._fill
                pg.strokeWidth = 0
                pg.strokeColor = self._fill
                d.add(pg)
            b = self._border
            d.drawOn(c, b, b)        
            return c

else:
    # if reportlab was not available, define a dummy
    class ChartGen_pdf():
        def __init__(*args, **kwargs):
            '''
            ChartGen_pdf is an extra that requires reportlab to work, but
            reportlab was not installed during import. So this is a dummy that
            will only raise an ImportError

            Try ::
                
                pip install opticiq[extras]
            '''
            raise ImportError('missing optional reportlab, try "pip install opticiq[extras]"')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Feb 25 14:26:02 2023

@author: chris
"""


__version__ = "0.0.5"
__status__ = "Alpha"

from .bg import *
from .edge import *
from .grad import *
from .peak import *
from .recipes import *
from .roi import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Regions has a lot of variability to bear in mind.
1. Number of ROI per image (1 to 1000s)
2. Need for ROI boundaries to evolve as analysis progresses from quick moments
    to fitting or other. Or not evolve.
3. Need to downselect POI/ROI at some point in an analysis. Or not.
4. Need to setup fixed ROI for production testing where expected
    result is fixed and auto is not wanted.
5. Must assume ROI are ragged (3d array is bad, also slower)
6. Need to avoid interference between ROI_i mask and any other ROI, or not

Created on Wed Mar 29 09:56:52 2023

@author: chris
"""

from abc import ABC as _ABC, abstractmethod as _abmeth
from matplotlib import pyplot as _plt
import numpy as _np
from skimage.measure import label as _label


def get_rslices(j1, j2, i1, i2):
    '''
    Parameters
    ----------
    j1, j2, i1, i2 : list or array of int
        Indices that define slices

    Returns
    -------
    rslices :  list of (jslice, islice)
        This is the format that rslices needs to follow.
        [(slice(j1[0]:j2[0]), slice(i1[0]:i2[0])),
         (slice(j1[0]:j2[1]), slice(i1[0]:i2[1]))...]
    '''
    Npoi = len(j1)
    rslices = [None] * Npoi
    for k in range(Npoi):
        jslice = slice(j1[k], j2[k])
        islice = slice(i1[k], i2[k])
        rslices[k] = (jslice, islice)
    return rslices


class quasiList_ABC(_ABC):
    '''
    Inheritable ABC with the purpose of making callables appear and behave
    like a list. Requires a child class to define ._select(k) and
    ._downselect(bool0, bool1, etc)

    The purpose of this is to delay execution, esp. for multi-threading when
    the number of ROI exceeds the number of cores.
    '''
    def __getitem__(self, indices):
        if _np.isscalar(indices):
            return self._select(indices)
        elif type(indices) is tuple:
            raise ValueError()
        else:
            return self._downselect(indices)

    def aslist(self):
        '''
        Executes _select(k) for each k, returning an actual list. This may
        save time at the expense of memory, if a given [k] will be retrieved
        multiple times.
        '''
        return [self._select(k) for k in range(len(self))]

    def __iter__(self):
        for k in range(len(self)):
            yield self._select(k)

    @_abmeth
    def __len__(self):
        pass

    @_abmeth
    def _select(self, k):
        pass

    @_abmeth
    def _downselect(self, idx):
        pass


class rmasks_fromlabels(quasiList_ABC):
    def __init__(self, rslices, labels, idx):
        '''
        Creates a list-like (quasiList_ABC) set of masks, where each
        mask corresponds to exactly one label, even if the labels are close
        together.

        Parameters
        ----------
        rslices : list of (jslice, islice)
            See get_rslices() for formatting.
        labels : 2d array of int
            See return value of skimage.measure.label
        idx : int, list, or array
            
        Returns
        -------
        rmasks : list-like
            Appears to be a list of 2d masks, but actually makes internal
            calls during member access. See quasiList_ABC
        '''
        self._idx = list(range(1, idx+1)) if type(idx) is int else idx
        assert len(self._idx) == len(rslices), 'idx is inconsistent with len(rslices)'
        self._labels = labels
        self._rslices = rslices

    def __len__(self):
        return len(self._idx)

    def _select(self, k):
        label = self._idx[k]
        jslice, islice = self._rslices[k]
        # note that the slice executes before the equality
        mask = self._labels[jslice, islice] == label
        return mask

    def _downselect(self, isvalid):
        idx = self._idx[isvalid]
        rslices = self._rslices[isvalid]
        return rmasks_fromlabels(rslices, self._labels, idx)


class rmasks_xyellipse(quasiList_ABC):
    def __init__(self, rslices, poi, xrad, yrad, x, y):
        '''
        Creates a list-like (quasiList_ABC) set of masks, where each
        mask is defined by elliptical radii xrad, yrad.

        Parameters
        ----------
        rslices : list of (jslice, islice)
            See get_rslices() for formatting.
        poi : array([[j0, i0], [j1, i1], ...])
            Points-of-Interest (same as peaks from
            skimage.feature.peak_local_max)
        xrad : array-like (often int type)
            semi-width of ROI
        yrad : array-like (often int type)
            semi-height of ROI
        x, y : 2d array
            global pixel coordinates

        Returns
        -------
        rmasks : list-like
            Appears to be a list of 2d masks, but actually makes internal
            calls during member access. See quasiList_ABC
        '''
        # also need x, y as inputs, though almost silly
        self._rslices = rslices
        self._poi = poi
        n = len(poi)
        self._xrad = xrad * _np.ones(n)
        self._yrad = yrad * _np.ones(n)
        self._x = x
        self._y = y

    def __len__(self):
        return len(self._poi)

    def _select(self, k):
        point = self._poi[k]
        cy, cx = tuple(point)
        jslice, islice = self._rslices[k]
        x = self._x[jslice, islice]
        y = self._y[jslice, islice]
        mask = (((x - cx)/self._xrad[k])**2
                + ((y - cy)/self._yrad[k])**2) <= 1
        return mask

    def _downselect(self, isvalid):
        poi = self._poi[isvalid]
        xrad = self._xrad[isvalid]
        yrad = self._yrad[isvalid]
        rslices = self._rslices[isvalid]
        return rmasks_xyellipse(
            rslices, poi, xrad, yrad, self._x, self._y)


class Regions():
    def __init__(self, rslices, rmasks=None):
        if rmasks is not None:
            assert len(rslices) == len(rmasks), 'mismatch length'
        self.rslices = rslices
        self.rmasks = rmasks

    def __len__(self):
        return len(self.rslices)

    def region_I(self, k, I):
        jslice, islice = self.rslices[k]
        val = I[jslice, islice]
        return val if (self.rmasks is None) else val * self.rmasks[k]

    def region_imG(self, k, imG, keys=None):
        '''
        Parameters
        ----------
        k : int
            which region of interest to select
        imG : dict of 2d arrays
            Normally, the original image array and its gradients, possibly
            including relevant masks. All 2d arrays are the same dimension as
            the original image
        keys : None or list of str, optional
            Override keys of imG. If None (default), all keys of imG are used.

        Returns
        -------
        imG_k : dict of 2d arrays
            Each key, val is sliced from imgrad using self.slices[k]    
        '''
        jslice, islice = self.rslices[k]
        imG_k = {}
        if keys is None:
            keys = imG.keys()
        for key in keys:
            imG_k[key] = self.region_I(k, imG[key])
        return imG_k

    def downselect(self, isvalid):
        rslices = self.rslices[isvalid]
        rmasks = None if (self.rmasks is None) else self.rmasks[isvalid]
        return Regions(rslices, rmasks)

    def plot_k(self, k, imG, keys=['ones', 'I0', 'I1'], axes=None):
        '''
        Make a subplot series for a single region.

        Parameters
        ----------
        k : int
            Select which region to plot/
        imG : dict of 2d arrays
            Needs to contain keys. 2d arrays of the same size as the original
            image, will be sliced by self.region_k(...)
        keys : list of str, optional
            Select the keys in imG to be plotted.
            The default is ['ones', 'I0', 'I1'].
        axes : tuple of pyplot Axes
            Optional. Provide axes (matching length of keys), i.e. in order to
            customize panel construction.

        Returns
        -------
        axes : tuple of pyplot Axes
            Useful for customizing axes
        '''
        Nplot = len(keys)
        if Nplot < 1: return
        if axes is None:
            _, axes = _plt.subplots(1, Nplot, layout='tight')
        _plt.figure()
        imG_k = self.region_imG(k, imG, keys=keys)
        for i in range(Nplot):
            key = keys[i]
            ax = axes[i]
            ax.imshow(imG_k[key], cmap='gray')
            ax.set_title(key)
        return axes

    @classmethod
    def from_mask(cls, mask, *args, min_area=1):
        '''
        Signatures::
    
            regions, rmasks = Regions.from_mask(mask, imG)
            regions, rmasks = Regions.from_mask(mask, x, y)
            regions, rmasks = Regions.from_mask(mask)

        Parameters
        ----------
        mask : 2d array of bool
            mask (same dimensions as original image) to be factored into Regions
        x, y : array of int, optional
            global pixel coordinates
        imG : dict of arrays, optional
            requires imG['x'] and imG['y']
        min_area : int, optional
            Any ROI of area < min_area will be excluded. The default is 1.
        '''
        if len(args) == 0:
            raise NotImplementedError()
        elif len(args) == 1:
            x = args[0]['x']
            y = args[0]['y']
        elif len(args) == 2:
            x, y = args
        else:
            raise ValueError('wrong number of args')
        label, n = _label(mask, return_num=True)
        j1 = []
        j2 = []
        i1 = []
        i2 = []
        ind = []
        for k in range(1, n+1):
            mask_k = label==k
            area = _np.sum(mask_k)
            if area >= min_area:
                ind.append(k)
                x_k = x[mask_k]
                y_k = y[mask_k]
                i1.append(_np.min(x_k))
                i2.append(_np.max(x_k))
                j1.append(_np.min(y_k))
                j2.append(_np.max(y_k))
        rslices = get_rslices(_np.array(j1), _np.array(j2),
                              _np.array(i1), _np.array(i2))
        rmasks = rmasks_fromlabels(rslices, label, ind)
        return cls(rslices, rmasks)

    @classmethod
    def from_poi(cls, poi, xrad, yrad, *args):
        '''
        Signatures::
            
            Regions.from_poi(poi, xrad, yrad, Ishape)
            Regions.from_poi(poi, xrad, yrad, x, y)
            Regions.from_poi(poi, xrad, yrad, imG)

        Get frame boundaries (e.g. for slicing frames at POI) using semi-width
        and semi-height, but enforce boundaries around Ishape.

        Parameters
        ----------
        poi : array([[j0, i0], [j1, i1], ...])
            Points-of-Interest (same as peaks from
            skimage.feature.peak_local_max)
        xrad : array-like (often int type)
            semi-width of ROI
        yrad : array-like (often int type)
            semi-height of ROI
        Ishape : (nj, ni)
            shape of image containing all POI, i.e. I.shape
        x, y : 2d array
            global pixel coordinates
        '''
        if len(args) == 1:
            if type(args[0]) is tuple:
                Ishape = args[0]
                #raise NotImplementedError()
            else:
                x, y = args[0]['x'], args[0]['y']
                Ishape = x.shape
        elif len(args) == 2:
            x, y = args
            Ishape = x.shape
        ny, nx = Ishape
        j = poi[:, 0]
        i = poi[:, 1]
        j1 = _np.maximum(0, j - yrad + .5).astype(int)
        j2 = _np.minimum(ny - 1, j + yrad + 1.5).astype(int)
        i1 = _np.maximum(0, i - xrad + .5).astype(int)
        i2 = _np.minimum(nx - 1, i + xrad + 1.5).astype(int)
        rslices = get_rslices(_np.array(j1), _np.array(j2),
                              _np.array(i1), _np.array(i2))
        rmasks = rmasks_xyellipse(rslices, poi, xrad, yrad, x, y)
        return cls(rslices, rmasks)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec 28 20:09:37 2023

@author: chris
"""

import numpy as _np
from scipy import ndimage as _ndimage


def genstarfield(res, nstar, sigma, magrange=[0, 4], E=1000, dark=1):
    '''
    Generate a blurry star field in the raw (array of float, without Poisson noise)

    Parameters
    ----------
    res : int or [int, int]
        resolution of image panel
    nstar : int
        number of stars in field
    sigma : float or [float, float]
        radius of Gaussian blur (see scipy.ndimage.gaussian_filter)
    magrange : [float, float], optional
        Magnitude range of stars in the field, where 1 is defined by E.
        Each star will have a randomly selected energy in this range.
        The default is [0, 4].
    E : float, optional
        Signal energy level, in counts, of Mag 0 star. The default is 1000.
    dark : float, optional
        Bias level, in counts, of darkness. The default is 1.

    Returns
    -------
    image_raw : 2d array of float
        Image of star field in floating point array format
    '''
    res *= _np.array([1, 1], dtype=int)
    hres = res[0]
    vres = res[1]
    #x, y = _np.meshgrid(_np.arange(res[0]), _np.arange(res[1]))
    canvas = dark*_np.ones((vres, hres))
    for i in range(nstar):
        x = _np.random.randint(0, hres)
        y = _np.random.randint(0, vres)
        mag = _np.random.randint(magrange[0], magrange[1]+1)
        canvas[y, x] = E * 100**(mag/5)
    image_raw = _ndimage.gaussian_filter(canvas, sigma)
    return image_raw



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Feb 25 14:31:36 2023

@author: chris
"""

import numpy as _np
from PIL import Image as _Image, ImageDraw as _ImageDraw
from sympy import Point as _Point, Polygon as _Polygon
from sympy.geometry.util import convex_hull as _convex_hull
try:
    from itertools import pairwise as _pairwise
except:
    from more_itertools import pairwise as _pairwise

from ..math import _rotate_vertices, _offset_vertices


def _naming_grid(frame, p, tile=True, unit=1):
    '''
    Naming string to describe pattern tiling
    '''
    p = _np.ones(2, dtype=int) * p
    if tile:
        d = p
    else:
        d = frame / p
    d = d / unit
    if _np.allclose(0, _np.mod(p, 1)):
        naming = ('%d' % d[0]) if _np.isclose(d[0], d[1]) else ('%dx%d' % (d[0], d[1]))
    else:
        naming = ('%0.2f' % d[0]) if _np.isclose(d[0], d[1]) else ('%0.2fx%0.2f' % (d[0], d[1]))
    return naming


def pattern_grid(frame, p, tile=True):
    '''
    Low-level utility common to all pattern generators, to control how
    patterns are tiled within a frame. Provides a tile mode and a stretch mode.

    E.g. pattern gen application may iterate pairwise over the grid::
        
        from itertools import pairwise
        xp, yp, _, _ = pattern_grid(frame, p, tile=tile)
        for x1, x2 in pairwise(xp):
            for y1, y2 in pairwise(yp):
                # x1, x2, y1, y2 are corners of the tile to draw a pattern

    Parameters
    ----------
    frame : [width, height]
        dimensions of the frame
    p : float or [float, float]
        (tile) - interval between patterns, (stretch) - number of patterns
    tile : bool, optional
        (True) - tile at even interval ; (False)- stretch pattern to fit the frame

    Returns
    -------
    xp, yp : 1d arrays
        Coordinates of points which define the corners of pattern tiles
        (not the centers of pattern tiles). Tiling always starts from the
        center of the frame, i.e. so that corners may clip non-integer tiles.
        ... And that means that non-integer framing can lead to corner points
        outside of the frame.
    d : [float, float]
        interval spacing between tiles
    N : [int, int]
        number of tiles
    '''
    p = _np.ones(2, dtype=int) * p
    frame = _np.ones(2, dtype=int) * frame
    centers = frame / 2
    if tile:
        d = p
        N = _np.ceil(frame / p)
    else:
        d = frame / p
        N = _np.ceil(p)
    corner1 = centers - (d * N) / 2
    corner2 = centers + (d * N) / 2
    xp = _np.linspace(corner1[0], corner2[0], int(N[0] + 1))
    yp = _np.linspace(corner1[1], corner2[1], int(N[1] + 1))
    return xp, yp, d, N


def limit_grid2frame(xp, yp, frame):
    xp = _np.minimum(xp, frame[0])
    xp = _np.maximum(xp, 0)
    yp = _np.minimum(yp, frame[1])
    yp = _np.maximum(yp, 0)
    return xp, yp


def limit_polygon2frame(frame, vertices):
    '''
    Some pattern generators may want to use a filter to clip polygons if they
    fall outside the frame.
    '''
    pgon_points = [_Point(v) for v in vertices]
    pgon_poly = _Polygon(*tuple(pgon_points))
    if pgon_poly.area < 0:
        # reverse polygons if the path area is - (CW order of vertices)
        pgon_poly = _Polygon(*tuple(reversed(pgon_points)))
        print('after reversal ', pgon_poly)
    w, h = tuple(frame)
    frame_points = [_Point(v) for v in [(0,0), (w, 0), (w, h), (0, h)]]
    frame_poly = _Polygon(*tuple(frame_points))
    # get all points enclosed in each other
    enclosed = []
    for point in pgon_points:
        if frame_poly.encloses(point):
            enclosed.append(point)
    for point in frame_points:
        if pgon_poly.encloses(point):
            enclosed.append(point)
    # intersect may be empty, a point, or a line (it's not a polygon)
    intersect = list(frame_poly.intersection(pgon_poly))
    # the answer is the convex hull of all enclosed points plus the intersect
    hull = _convex_hull(*tuple(enclosed + intersect))
    # we do want to turn it back into a list of tuples though
    return [tuple(p) for p in hull.vertices]
    #return hull


def itr_pointgrid(frame, p, **kwargs):
    '''
    Parameters
    ----------
    see pattern_grid

    Yields
    ------
    x, y : float, float
        x, y point to feed to a plotter

    Suppresses points outside the frame.
    '''
    xp, yp, _, _ = pattern_grid(frame, p, **kwargs)
    for x1, x2 in _pairwise(xp):
        for y1, y2 in _pairwise(yp):
            x, y = (x1 + x2) / 2, (y1 + y2) / 2
            if x >= 0 and x < frame[0] and y >= 0 and y < frame[1]:
                yield x, y


def itr_checkerboard(frame, p, tile=True, even=True):
    '''
    Parameters
    ----------
    see pattern_grid

    Yields
    ------
    x1, y1, x2, y2 : 
        upper-left and lower-right rectangle corners to feed to plotter

    Clips checkers to frame if needed.
    '''
    xp, yp, _, _ = pattern_grid(frame, p, tile=tile)
    xp, yp = limit_grid2frame(xp, yp, frame)
    Nh = len(xp) - 1
    Nv = len(yp) - 1
    mod = 0 if even else 1
    for col in range(Nh):
        for row in range(Nv):
            if _np.mod(row + col, 2) == mod:
                # yield alternating patterns
                yield xp[col], yp[row], xp[col + 1], yp[row + 1]


def itr_slant_rectangles(frame, p, tile=True, slant=5, fill_factor=.5):
    '''
    Parameters
    ----------
    frame, p, tile :
        see pattern_grid
    slant : float, optional
        angle in degrees to slant the squares (default = 5 degrees)
    fill_factor : float 0 to ~.95 (default = 0.5)
        low fill factor means more space between slanted squares

    Yields
    ------
    vertices :
        vertices [(x0, y0), (x1, y1), ...] to feed to a polygon plotter

    Does not clip polygon vertices to frame (see limit_polygon2frame).
    '''
    xp, yp, d, _ = pattern_grid(frame, p, tile=tile)
    half_w, half_h = tuple(d * fill_factor / 2)
    vertices0 = [(-half_w, -half_h), (half_w, -half_h),
                 (half_w, half_h), (-half_w, half_h)]
    vertices0 = _rotate_vertices(vertices0, slant)
    for x1, x2 in _pairwise(xp):
        for y1, y2 in _pairwise(yp):
            cx, cy = (x1 + x2) / 2, (y1 + y2) / 2
            vertices = _offset_vertices(cx, cy, vertices0)
            yield vertices


class ChartGen_img():
    '''
    Raster image chart generator for optical testing.

    Example::
        
        igen = ChartGetn_img([1600, 1200], cr=12)
        im, name = igen.checkerboard(20)
        im.save(name + '.png')
    '''
    def __init__(self, res, mode='W', cr=None, fg=255, bg=0):
        '''
        Parameters
        ----------
        res : [hres, vres]
            resolution in pixels
        [mode] : str
            color mode (default='W') - 'W', 'R', 'G', 'B'
        [contrast] : float, None means ignore
            contrast ratio; bg=round(fg/(1 + contrast))
        [fg] : int 0-255
            foreground (default 255)
        [bg] : int 0-255
            background (default 0)
        '''
        mode = mode.upper()
        if cr is not None:
            bg = int(fg / (1 + cr) + .5)
        self._basename = '%dx%d' % tuple(res) + '%s_%don%d_' % (mode, fg, bg)
        if mode == 'W':
            imgmode = 'L'
            imgfg = (fg,)
            imgbg = (bg,)
        elif mode == 'R':
            imgmode = 'RGB'
            imgfg = (fg, 0, 0)
            imgbg = (bg, 0, 0)
        elif mode.lower() == 'G':
            imgmode = 'RGB'
            imgfg = (0, fg, 0)
            imgbg = (0, bg, 0)
        elif mode.lower() == 'B':
            imgmode = 'RGB'
            imgfg = (0, 0, fg)
            imgbg = (0, 0, bg)
        else:
            raise ValueError('unknown mode "%s"' % mode)
        self.res = res
        self.imgfg = imgfg
        self.imgbg = imgbg
        self.imgmode = imgmode

    def _get_blank(self):
        return _Image.new(self.imgmode, tuple(self.res), self.imgbg)

    def pointgrid(self, p, **kwargs):
        '''
        Parameters
        ----------
        p : float or [float, float]
            (tile) - interval between patterns in unit ;
            (stretch) - number of patterns
        tile : bool, optional
            (True) - tile at even interval ; (False) - stretch pattern to fit the frame

        Returns
        -------
        img : PIL.Image
            test chart; grid of points
        name : str
            Name (recommended filename without extension)
        '''
        img = self._get_blank()
        draw = _ImageDraw.Draw(img)
        for xy in itr_pointgrid(self.res, p, **kwargs):
            draw.point(xy, fill=self.imgfg)
        name = self._basename + 'pointgrid_'
        name += _naming_grid(self.res, p, **kwargs) + 'pix'
        return img, name

    def checkerboard(self, p, **kwargs):
        '''
        Parameters
        ----------
        p : float or [float, float]
            (tile) - interval between patterns in unit ;
            (stretch) - number of patterns
        tile : bool, optional
            (True) - tile at even interval ; (False) - stretch pattern to fit the frame

        Returns
        -------
        img : PIL.Image
            test chart; checkerboard
        name : str
            Name (recommended filename without extension)
        '''
        img = self._get_blank()
        draw = _ImageDraw.Draw(img)
        for x1, y1, x2, y2 in itr_checkerboard(self.res, p, **kwargs):
            draw.rectangle((x1, y1, x2 - 1, y2 - 1),
                           fill=self.imgfg, outline=self.imgfg)
        name = self._basename + 'checker_'
        name += _naming_grid(self.res, p, **kwargs) + 'pix'
        return img, name

    def slant_rectangles(self, p, tile=True, slant=5, fill_factor=0.5):
        '''
        Parameters
        ----------
        p : float or [float, float]
            (tile) - interval between patterns in unit,
            or (stretch) - number of patterns
        tile : bool, optional
            (True) - tile at even interval ; (False) - stretch pattern to fit the frame
        slant : float, optional
            angle in degrees to slant the squares (default = 5 degrees)
        fill_factor : float 0 to ~.95 (default = 0.5)
            low fill factor means more space between slanted squares

        Returns
        -------
        img : PIL.Image
            test chart; slanted rectangles, use for slant-edge SFR
        name : str
            Name (recommended filename without extension)
        '''
        img = self._get_blank()
        draw = _ImageDraw.Draw(img)
        for vertices in itr_slant_rectangles(
                self.res, p, tile=tile, slant=slant, fill_factor=fill_factor):
            draw.polygon(vertices, fill=self.imgfg, outline=self.imgfg)
        name = self._basename + 'slantrec_'
        name += _naming_grid(self.res, p, tile=tile) + 'pix'
        return img, name

    def checkers_and_points(self, p, tile=True):
        '''
        '''
        img = self._get_blank()
        draw = _ImageDraw.Draw(img)
        for x1, y1, x2, y2 in itr_checkerboard(self.res, p, tile=tile,
                                               even=True):
            draw.rectangle((x1, y1, x2 - 1, y2 - 1),
                           fill=self.imgfg, outline=self.imgfg)
        for x1, y1, x2, y2 in itr_checkerboard(self.res, p, tile=tile,
                                               even=False):
            draw.point(((x1 + x2)/2, (y1 + y2)/2), fill=self.imgfg)
        name = self._basename + 'checkerpoints_'
        name += _naming_grid(self.res, p, tile=tile) + 'pix'
        return img, name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 10 18:59:25 2023

@author: chris
"""

import numpy as _np
from scipy import ndimage as _ndimage
from skimage import morphology as _morph


def mask(f, threshold=0.1, f0=0):
    '''mask = (f - f0) / np.max(f) > threshold'''
    return (f - f0) / _np.max(f) > threshold


def maskedges(imgrad, threshold):
    imgrad['maskedges'] = mask(
        imgrad['I_r'], threshold)


def maskdark(imgrad):
    mid = _np.median(imgrad['I1'].flatten()[imgrad['maskedges'].flatten()])
    maskdark0 = imgrad['I1'] < mid
    imgrad['maskdark'] = _np.logical_and(
        maskdark0,
        _np.logical_not(imgrad['maskedges']))


def maskedges_hvx(imgrad, threshold):
    mask_v_abs = mask(_np.abs(imgrad['I_x']), threshold)
    mask_h_abs = mask(_np.abs(imgrad['I_y']), threshold)
    mask_v_pos0 = mask(imgrad['I_x'], threshold)
    mask_v_neg0 = mask(-imgrad['I_x'], threshold)
    mask_h_pos0 = mask(imgrad['I_y'], threshold)
    mask_h_neg0 = mask(-imgrad['I_y'], threshold)
    masks = [_np.logical_and(mask_v_pos0, _np.logical_not(mask_h_abs)),
             _np.logical_and(mask_v_neg0, _np.logical_not(mask_h_abs)),
             _np.logical_and(mask_h_pos0, _np.logical_not(mask_v_abs)),
             _np.logical_and(mask_h_neg0, _np.logical_not(mask_v_abs))]
    for i in range(4):
        # erosion of dilation can cleanup if there's noise in a mask
        masks[i] = _morph.erosion(_morph.dilation(masks[i]))
    imgrad['mask_v_pos'] = masks[0]
    imgrad['mask_v_neg'] = masks[1]
    imgrad['mask_h_pos'] = masks[2]
    imgrad['mask_h_neg'] = masks[3]


def maskridges_hvx(imgrad, threshold):
    mask_v0 = mask(-imgrad['I_xx'], threshold)
    mask_h0 = mask(-imgrad['I_yy'], threshold)
    imgrad['mask_v'] = _np.logical_and(mask_v0, _np.logical_not(mask_h0))
    imgrad['mask_h'] = _np.logical_and(mask_h0, _np.logical_not(mask_v0))


def imageGradients(I0, sigma, require=[], threshold=None):
    '''
    Preprocessing steps potentially applicable to every image quality test.

    Given a raw image, apply Gaussian blur, make derivatives, package results
    as a dict of 2d arrays. The keys of the dict follow a naming convention,
    see below.

    Parameters
    ----------
    I0 : 2d array
        Original image
    sigma : float
        Sigma (pixels) to blur image before derivatives
    require : list of str
        Each str in list should be a standard name of an array (above).
        Dependencies can be resolved automatically, so ['D_hessian'] will
        also call all the first and second order derivatives.
        The default is [], which will only return the blurred image.
    threshold : TYPE, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    imgrad : dict of 2d arrays
        All results required or implied.

    Names of Arrays
    ---------------
    | x, y : global coordinates (column, row) of pixels
    | ones : 1s for each pixel, initially. Masks leave a footprint
    | I0 : original image
    | I1 : blurred image
    | I_x : dI1/dx
    | I_y : dI1/dy
    | I_r : sqrt(I_x**2 + I_y**2) aka magnitude of slope
    | I_xx : d2I1/dxdx
    | I_xy : d2I1/dxdy
    | I_yx : d2I1/dydx
    | I_yy : d2I1/dydy
    | curve : I_xx + I_yy aka curvature, which goes to 0 at saddle point
    | D_hessian : I_xx * I_yy - I_xy * I_yx
    | maskedges : I_r / max(I_r) < threshold
    | mask_h_abs : abs(I_y)/max(abs(I_y)) < threshold
    | mask_v_abs : ... similar to mask_h_abs
    | mask_v_pos : I_x / max(I_x) < threshold and not mask_h_abs 
    | mask_v_neg, mask_h_pos, mask_h_get : .., similar to mask_v_pos
    '''
    # some logic finds implied parameters
    do_dhesse = 'D_hessian' in require
    do_curve = 'curve' in require
    do_d2 = (do_dhesse or do_curve or 'I_xx' in require)
    do_Ir = 'I_r' in require
    do_d1 = (do_d2 or do_Ir or 'I_x' in require or 'I_y' in require)
    # xframe and yframe indicate global coordinates of pixels
    ny, nx = I0.shape
    x, y = _np.meshgrid(range(nx), range(ny))
    # initial dataset is xframe, yframe, and original image
    imG = dict(I0=I0, x=x, y=y, ones=_np.ones(I0.shape))
    # add blurred image
    imG['I1'] = _ndimage.gaussian_filter(I0, sigma)
    if do_d1:
        # first derivatives
        imG['I_y'], imG['I_x'] = _np.gradient(imG['I1'])
    if do_Ir:
        imG['I_r'] = _np.sqrt(imG['I_y']**2 + imG['I_x']**2)
    if do_d2:
        # second derivatives
        imG['I_xy'], imG['I_xx'] = _np.gradient(imG['I_x'])
        imG['I_yy'], imG['I_yx'] = _np.gradient(imG['I_y'])
    if do_curve:
        imG['curve'] = imG['I_xx'] + imG['I_yy'] #+ self.I_xy + self.I_yx
    if do_dhesse:
        # Hessian determinant
        imG['D_hessian'] = imG['I_xx'] * imG['I_yy'] - imG['I_xy'] * imG['I_yx']
    return imG



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
opticiq.gen
===========
For generating pattern images used in image quality testing.

See ChartGen_pdf, ChartGen_img, ChartGen_gds

Created on Sat Feb 25 15:13:43 2023

@author: chris
"""

from .chart import *
from .extra import *
from .gauss import *
from .model import *
from .star import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Mar 11 20:10:29 2023

@author: chris
"""


class MTF_Analysis():
    '''
    MTF_Analysis can find optical Modulation Transfer Function (MTF) using the
    edges of slant-rectangle test images (recommended) or using the
    line-spread-function of cross-hair test images.

    Example::
        
        import numpy as np
        im = np.asarray(...)
        mtfa = MTF_Analysis(im)
        mtfa.edge2mtf()
    '''
    def __init__(self):
        raise NotImplementedError

    def lsf(self):
        pass

    def lsf2mtf(self):
        pass

    def edge(self):
        pass

    def edge2mtf(self):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
opticiq.opt subpackage contains very simple optical models such as FFT-->PSF.
These can be useful helpers, e.g. forward models can be good for code testing,
data fitting algoriths, or Monte Carlo methods.

Created on Sat Feb  1 11:39:20 2025

@author: chris
"""

from .fft import *
